from unittest.mock import Mock

import pytest
import pytest_asyncio

from cashews.backends.memory import Memory
from cashews.decorators.bloom import bloom, dual_bloom

pytestmark = pytest.mark.asyncio


@pytest_asyncio.fixture(
    name="backend",
    params=[
        "memory",
        pytest.param("redis", marks=pytest.mark.redis),
    ],
)
async def _cache(request, redis_dsn, backend_factory):
    if request.param == "diskcache":
        from cashews.backends.diskcache import DiskCache

        backend = await backend_factory(DiskCache, shards=0)
        yield backend
    elif request.param == "redis":
        from cashews.backends.redis import Redis

        yield await backend_factory(Redis, redis_dsn, hash_key=None)
    else:
        yield await backend_factory(Memory)


async def test_bloom_simple(backend):
    n = 100
    call = Mock()

    @bloom(backend=backend, name="name:{k}", false_positives=1, capacity=n)
    async def func(k):
        call(k)
        return k > (n / 2)

    for i in range(n):
        await func.set(i)

    for i in range(n):
        call.reset_mock()
        assert await func(i) is (i > (n / 2))
        if i > (n / 2):
            call.assert_called_with(i)
        else:
            call.assert_not_called()


async def test_bloom_simple_big_size(backend):
    n = 1_000_000
    call = Mock()

    @bloom(backend=backend, name="name:{k}", false_positives=1, capacity=n)
    async def func(k):
        call(k)
        return k > (n / 2)

    await func.set(900_000)
    await func.set(1_000)

    call.reset_mock()
    assert await func(900_000)
    call.assert_called_with(900_000)

    call.reset_mock()
    assert not await func(1_000)
    call.assert_not_called()


async def test_bloom_dual(backend):
    n = 100
    call = Mock()

    @dual_bloom(backend=backend, name="name:{k}", false=1, capacity=n)
    async def func(k):
        call(k)
        return k > (n / 2)

    for i in range(n):
        await func(i)

    call.reset_mock()
    for i in range(n):
        assert await func(i) is (i > (n / 2))
        call.assert_not_called()

    await func.delete(10)
    await func.delete(70)

    assert await func(10) is False
    call.assert_called_with(10)

    call.reset_mock()
    assert await func(70) is True
    call.assert_called_with(70)



import dataclasses
from collections import namedtuple
from datetime import date, datetime, timedelta
from decimal import Decimal

import pytest
import pytest_asyncio
from hypothesis import example, given, settings
from hypothesis import strategies as st

from cashews.serialize import PickleSerializerMixin, UnSecureDataError

pytestmark = pytest.mark.asyncio


@dataclasses.dataclass()
class TestDC:
    test: str
    _: int


NT = namedtuple("NT", ("test", "name"), defaults=[False, "test"])


class DummyCache:
    def __init__(self):
        self.store = {}

    async def set(self, key: str, value, *args, **kwargs) -> None:
        self.store[key] = value

    async def get(self, key: str, default=None, **kwargs):
        return self.store.get(key, default)

    async def delete(self, *keys):
        for key in keys:
            del self.store[key]

    async def get_many(self, *keys, default=None):
        return [self.store.get(key, default) for key in keys]

    async def set_many(self, pairs, expire=None):
        self.store.update(pairs)


class Cache(PickleSerializerMixin, DummyCache):
    pass


@pytest_asyncio.fixture(
    name="cache",
    params=[
        "default_md5",
        "default_sum",
        "default_sha256",
        pytest.param("redis_md5", marks=pytest.mark.redis),
        pytest.param("dill_sum", marks=pytest.mark.integration),
        pytest.param("sqlalchemy_sha1", marks=pytest.mark.integration),
    ],
)
async def _cache(request, redis_dsn):
    pickle_type, digestmod = request.param.split("_")
    if pickle_type == "redis":
        from cashews.backends.redis import Redis

        redis = Redis(redis_dsn, hash_key=b"test", safe=False, digestmod=digestmod)
        await redis.init()
        await redis.clear()
        return redis
    return Cache(hash_key="test", digestmod=digestmod, pickle_type=pickle_type)


@pytest.mark.parametrize(
    "value",
    (
        "test",
        b"test",
        "1_1",
        b"1_1",
        b"1_1_1",
        0,
        1,
        2,
        1.234,
        Decimal("1.001"),
        True,
        False,
        "",
        None,
        {"hi": True},
        TestDC(test="test", _=1),
    ),
)
async def test_serialize_simple_value(value, cache):
    await cache.set("key", value)
    await cache.set_many({"key1": value, "key2": value})
    assert await cache.get("key") == value
    assert await cache.get("key2") == value
    assert await cache.get_many("key1", "key2") == (value, value)


@pytest.mark.parametrize("value", (NT(name="test", test="lol"), NT()))
async def test_serialize_collections_value(value, cache):
    await cache.set("key", value)
    assert await cache.get("key") == value
    assert (await cache.get("key")).name == "test"


@pytest.mark.parametrize(
    "value",
    (
        datetime(year=2000, month=1, day=10, hour=10),
        timedelta(days=10),
        date(year=2020, month=12, day=31),
    ),
)
async def test_serialize_dates_value(value, cache):
    await cache.set("key", value)
    assert await cache.get("key") == value


@pytest.mark.parametrize(
    "value",
    (
        ["test", "to"],
        (TestDC("hay", _=2),),
        {"test"},
        [(1, 2), (3, 4)],
        {1, 2, 5},
        [{"test": True}],
    ),
)
async def test_serialize_array_value(value, cache):
    await cache.set("key", value)
    assert list(await cache.get("key")) == list(value)
    assert type(await cache.get("key")) == type(value)


@pytest.mark.parametrize(
    "value",
    (
        ["test", b"to"],
        (TestDC("hay", _=2), Decimal("10.1")),
        {"test", 1},
        [],
        [{1: True}, Decimal("0.1")],
    ),
)
async def test_serialize_array_diff_value(value, cache):
    await cache.set("key", value)
    assert list(await cache.get("key")) == list(value)
    assert type(await cache.get("key")) == type(value)


@pytest.mark.parametrize(
    "value",
    (
        b"_cos\nsystem\n(S'echo hello world'\ntR.",
        b":_cos\nsystem\n(S'echo hello world'\ntR.",
        b"md5:_cos\nsystem\n(S'echo hello world'\ntR.",
        b"sha1:_cos\nsystem\n(S'echo hello world'\ntR.",
        b"__cos\nsystem\n(S'echo hello world'\ntR.",
    ),
)
async def test_unsecure_value(value, cache):
    await cache.set_raw("key", value)
    with pytest.raises(UnSecureDataError):
        assert not await cache.get("key")


async def test_unsecure_value_many(cache):
    await cache.set_raw("key", b"_cos\nsystem\n(S'echo hello world'\ntR.")
    with pytest.raises(UnSecureDataError):
        await cache.get_many("key")


async def test_no_value(cache):
    assert await cache.get("key") is None


async def test_replace_values(cache):
    await cache.set("key", "key")
    await cache.set_raw("replace", await cache.get_raw("key"))

    with pytest.raises(UnSecureDataError):
        await cache.get("replace")


async def test_pickle_error_value(cache):
    await cache.set_raw(
        "key",
        cache._gen_sign("key", b"no_pickle_data", cache._digestmod) + b"_" + b"no_pickle_data",
    )
    assert await cache.get("key") is None


async def test_set_no_ser(cache):
    empty = object()
    await cache.set("key_e", empty)


Schema = None


async def test_data_change(cache):
    global Schema

    @dataclasses.dataclass
    class Schema:
        test: str

    await cache.set("key", Schema("test"))

    @dataclasses.dataclass
    class Schema:
        name: str

    # Check without data change
    await cache.set("key2", Schema("test"))
    value = await cache.get("key2")
    assert value.name == "test"

    # IN key class of data have changed so it should be invalid
    value = await cache.get("key")
    assert value is None


async def test_get_set_raw(cache):
    await cache.set_raw("key", b"test")
    assert await cache.get_raw("key") == b"test"


@given(key=st.text(), value=st.characters())
@settings(max_examples=500)
@example(key="_key:_!@#$%^&*()", value='_value:_!@_#$%^&:*(?)".,4й')
async def test_no_hash(key, value):
    cache = Cache()
    await cache.set(key, value)
    assert await cache.get(key) == value


async def test_cache_from_hash_to_no_hash():
    val = Decimal("10.2")
    cache = Cache(hash_key="test")
    await cache.set("key", val)

    assert await cache.get("key") == val
    cache_no_hash = Cache()
    cache_no_hash.store = cache.store
    assert await cache_no_hash.get("key") == val


async def test_cache_from_no_hash_to_hash():
    val = Decimal("10.2")
    cache = Cache()
    await cache.set("key", val)

    assert await cache.get("key") == val
    cache_hash = Cache(hash_key="test")
    cache_hash.store = cache.store

    assert await cache_hash.get("key") is None



import asyncio
import logging
from datetime import datetime, timedelta
from functools import wraps
from typing import Optional

from ..._typing import TTL, AsyncCallable_T, CallableCacheCondition, Decorator
from ...backends.interface import _BackendInterface
from ...formatter import register_template
from ...key import get_cache_key, get_cache_key_template
from ...ttl import ttl_to_seconds
from .defaults import _empty, context_cache_detect

__all__ = ("early",)


logger = logging.getLogger(__name__)
_LOCK_SUFFIX = ":lock"


def early(
    backend: _BackendInterface,
    ttl: TTL,
    key: Optional[str] = None,
    early_ttl: Optional[TTL] = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "early",
) -> Decorator:
    """
    Cache strategy that try to solve Cache stampede problem (https://en.wikipedia.org/wiki/Cache_stampede),
    With hot cache recalculate a result in background near expiration time
    Warning Not good at cold cache

    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param early_ttl: duration in seconds to expire results
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'early'
    """

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix + ":v2")
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs)
            if early_ttl is None:
                _early_ttl = _ttl * 0.33
            else:
                _early_ttl = ttl_to_seconds(early_ttl, *args, **kwargs)

            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            cached = await backend.get(_cache_key, default=_empty)
            if cached is not _empty:
                context_cache_detect._set(
                    _cache_key,
                    ttl=_ttl,
                    early_ttl=_early_ttl,
                    name="early",
                    template=_key_template,
                )
                early_expire_at, result = cached
                if early_expire_at <= datetime.utcnow() and await backend.set(
                    _cache_key + _LOCK_SUFFIX, "1", expire=_early_ttl, exist=False
                ):
                    logger.info(
                        "Recalculate cache for %s (exp_at %s)",
                        _cache_key,
                        early_expire_at,
                    )
                    asyncio.create_task(
                        _get_result_for_early(
                            backend,
                            func,
                            args,
                            kwargs,
                            _cache_key,
                            _ttl,
                            _early_ttl,
                            condition,
                            unlock=True,
                        )
                    )
                return result
            return await _get_result_for_early(
                backend,
                func,
                args,
                kwargs,
                _cache_key,
                _ttl,
                _early_ttl,
                condition,
            )

        return _wrap

    return _decor


async def _get_result_for_early(
    backend: _BackendInterface, func, args, kwargs, key, ttl: int, early_ttl: int, condition, unlock=False
):
    try:
        result = await func(*args, **kwargs)
        if condition(result, args, kwargs, key):
            early_expire_at = datetime.utcnow() + timedelta(seconds=early_ttl)
            await backend.set(key, [early_expire_at, result], expire=ttl)
        return result
    finally:
        if unlock:
            asyncio.create_task(backend.delete(key + _LOCK_SUFFIX))



import asyncio
import random

import pytest

from cashews import Cache
from cashews.validation import invalidate_func, invalidate_further

pytestmark = pytest.mark.asyncio


@pytest.fixture(name="cache")
def _cache():
    cache = Cache()
    cache.setup("mem://")
    return cache


async def test_invalidate_func(cache: Cache):
    @cache(ttl=1)
    async def func(arg):
        return random.random()

    first_call = await func("test")
    await asyncio.sleep(0)

    assert first_call == await func("test")
    await invalidate_func(cache, func, kwargs={"arg": "test"})
    assert first_call != await func("test")


async def test_invalidate_func_wrong(cache: Cache):
    @cache(ttl=1)
    async def func(arg):
        return random.random()

    first_call = await func("test")
    await asyncio.sleep(0)

    assert first_call == await func("test")
    await invalidate_func(cache, func, kwargs={"arg": "test2"})
    assert first_call == await func("test")


async def test_invalidate_decor(cache: Cache):
    @cache(ttl=1)
    async def func(arg):
        return random.random()

    @cache.invalidate(func, args_map={"arg": "a"})
    async def func2(a):
        return random.random()

    first_call = await func("test")
    await asyncio.sleep(0)

    assert first_call == await func("test")
    await func2("test")
    await asyncio.sleep(0)
    assert first_call != await func("test")


async def test_invalidate_decor_str(cache: Cache):
    @cache(ttl=1, key="key:{arg}")
    async def func(arg):
        return random.random()

    @cache.invalidate("key:{arg}")
    async def func2(arg, default=True):
        return random.random()

    first_call = await func("test")
    await asyncio.sleep(0)

    assert first_call == await func("test")
    await func2("test")
    await asyncio.sleep(0)
    assert first_call != await func("test")


async def test_invalidate_decor_complicate(cache: Cache):
    @cache(ttl=1, key="arg-{arg}:{key}:{flag}")
    async def func(arg, key="test", flag=True, **kwargs):
        return random.random()

    @cache.invalidate(func, args_map={"arg": "a"}, defaults={"flag": True})
    async def func2(a):
        return random.random()

    first_call = await func("test")
    not_invalidate = await func("test", flag=False)
    second_call = await func("test", "key", flag=True)
    last_call = await func("test2", None)
    await asyncio.sleep(0)

    assert first_call != second_call != last_call
    assert first_call == await func("test")
    assert second_call == await func("test", "key")
    assert last_call == await func("test2", None)
    assert not_invalidate == await func("test", flag=False)

    await func2("test3")

    await asyncio.sleep(0)
    assert first_call == await func("test")
    assert second_call == await func("test", "key")
    assert last_call == await func("test2", None)
    assert not_invalidate == await func("test", flag=False)

    await func2("test")

    await asyncio.sleep(0)
    assert first_call != await func("test")
    assert second_call != await func("test", "key")
    assert last_call == await func("test2", None)
    assert not_invalidate == await func("test", flag=False)

    await func2("test2")

    await asyncio.sleep(0)
    assert first_call != await func("test")
    assert not_invalidate == await func("test", flag=False)


async def test_invalidate_further_decorator(cache):
    @cache(ttl=10000)
    async def func():
        return random.random()

    first = await func()
    assert await func() == first

    with invalidate_further():
        on_clear = await func()
    assert on_clear is not None
    assert on_clear != first  # the key was deleted
    second = await func()
    assert second != first
    assert await func() == second


async def test_invalidate_further_get(cache):

    await cache.set("key", "value")
    assert await cache.get("key") == "value"

    with invalidate_further():
        assert await cache.set("key", "value2")
        assert await cache.set("key2", "value2")
        assert await cache.get("key") is None

    await asyncio.sleep(0)
    assert await cache.get("key") is None
    assert await cache.get("key2") == "value2"
    assert await cache.set("key", "value3")
    assert await cache.get("key") == "value3"


async def test_invalidate_further_get_many(cache):

    await cache.set("key", "value")
    await cache.set("key2", "value2")

    assert await cache.get_many("key", "key2") == ("value", "value2")

    with invalidate_further():
        assert await cache.get_many("key") == (None,)

    await asyncio.sleep(0)
    assert await cache.get_many("key", "key2") == (None, "value2")


async def test_invalidate_further_get_match(cache):

    await cache.set("key1", "value1")
    await cache.set("key2", "value2")
    assert {k: v async for k, v in cache.get_match("key*")} == {"key1": "value1", "key2": "value2"}

    with invalidate_further():
        assert {k: v async for k, v in cache.get_match("key1*")} == {}

    await asyncio.sleep(0)
    assert {k: v async for k, v in cache.get_match("key*")} == {"key2": "value2"}



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:  # pragma: no cover
    from cashews.backends.interface import Backend


pytest_plugins = ["pytest_asyncio"]  # pylint: disable=invalid-name

try:
    import aiohttp
except ImportError:
    pass
else:
    pytest_plugins.append("aiohttp.pytest_plugin")

    del aiohttp


@pytest.fixture
def redis_dsn():
    host = os.getenv("REDIS_HOST", "")
    port = os.getenv("REDIS_PORT", "")
    return f"redis://{host}:{port}/"


@pytest.fixture
def backend_factory():
    async def factory(backend_cls: type[Backend], *args, **kwargs):
        backend = backend_cls(*args, **kwargs)
        await backend.init()
        await backend.clear()
        return backend

    return factory



from unittest.mock import Mock

import pytest

from cashews import Cache
from cashews.backends.memory import Memory
from cashews.helpers import add_prefix, all_keys_lower, memory_limit

pytestmark = pytest.mark.asyncio


@pytest.fixture(name="target")
def _target() -> Mock:
    class MemCache(Memory):
        pass

    return Mock(wraps=MemCache())


@pytest.fixture(name="cache")
def __cache(target):
    _cache = Cache()
    _cache._add_backend(target)
    return _cache


async def test_all_keys_lower(cache: Cache, target):
    cache._add_backend(target, (all_keys_lower(),))
    await cache.get(key="KEY")
    target.get.assert_called_once_with(key="key", default=None)

    await cache.set(key="KEY", value="value")
    target.set.assert_called_once_with(
        key="key",
        value="value",
        exist=None,
        expire=None,
    )
    await cache.set_many({"KEY": "value"})
    target.set_many.assert_called_once_with(
        pairs={"key": "value"},
        expire=None,
    )
    await cache.ping()
    target.ping.assert_called_once_with(message=b"PING")


async def test_memory_limit(cache: Cache, target):
    cache._add_backend(target, (memory_limit(min_bytes=52, max_bytes=75),))

    await cache.set(key="key", value="v")
    target.set.assert_not_called()

    await cache.set(key="key", value="v" * 31)
    target.set.assert_not_called()

    await cache.set(key="key", value="v" * 15)
    target.set.assert_called_once_with(
        key="key",
        value="v" * 15,
        exist=None,
        expire=None,
    )

    await cache.ping()
    target.ping.assert_called_once_with(message=b"PING")


async def test_add_prefix(cache: Cache, target):
    cache._add_backend(target, (add_prefix("prefix!"),))

    await cache.get(key="key")
    target.get.assert_called_once_with(key="prefix!key", default=None)

    await cache.set(key="key", value="value")
    target.set.assert_called_once_with(
        key="prefix!key",
        value="value",
        exist=None,
        expire=None,
    )
    await cache.ping()
    target.ping.assert_called_once_with(message=b"PING")


async def test_add_prefix_get_many(cache: Cache, target):
    cache._add_backend(target, (add_prefix("prefix!"),))
    await cache.get_many("key")
    target.get_many.assert_called_once_with("prefix!key")


async def test_add_prefix_set_many(cache: Cache, target):
    cache._add_backend(target, (add_prefix("prefix!"),))
    await cache.set_many({"key": "value"})
    target.set_many.assert_called_once_with(pairs={"prefix!key": "value"}, expire=None)


async def test_add_prefix_delete_match(cache: Cache, target):
    cache._add_backend(target, (add_prefix("prefix!"),))
    await cache.delete_match("key")
    target.delete_match.assert_called_once_with(pattern="prefix!key")



from typing import Any, Callable, Dict, Tuple, Type, Union
from urllib.parse import parse_qsl, urlparse

from .backends.interface import Backend
from .backends.memory import Memory
from .exceptions import BackendNotAvailableError

try:
    from .backends.client_side import BcastClientSide
    from .backends.redis import Redis
except ImportError:
    BcastClientSide, Redis = None, None

try:
    import diskcache  # noqa: F401
except ImportError:
    DiskCache = None
else:
    from .backends.diskcache import DiskCache


_NO_REDIS_ERROR = "Redis backend requires `redis` to be installed."
_CUSTOM_ERRORS = {
    "redis": _NO_REDIS_ERROR,
    "rediss": _NO_REDIS_ERROR,
    "disk": "Disk backend requires `diskcache` to be installed.",
}
BackendOrFabric = Union[Type[Backend], Callable[..., Backend]]
_BACKENDS: Dict[str, Tuple[BackendOrFabric, bool]] = {}


def register_backend(alias: str, backend_class: BackendOrFabric, pass_uri: bool = False):
    _BACKENDS[alias] = (backend_class, pass_uri)


def _redis_fabric(**params: Any) -> Union[Redis, BcastClientSide]:
    if params.pop("client_side", None):
        return BcastClientSide(**params)
    return Redis(**params)


register_backend("mem", Memory)
if Redis:
    register_backend("redis", _redis_fabric, pass_uri=True)
    register_backend("rediss", _redis_fabric, pass_uri=True)
if DiskCache:
    register_backend("disk", DiskCache)


def settings_url_parse(url: str) -> Tuple[BackendOrFabric, Dict[str, Any]]:
    parse_result = urlparse(url)
    params: Dict[str, Any] = dict(parse_qsl(parse_result.query))
    params = serialize_params(params)

    alias = parse_result.scheme
    if alias == "":
        return Memory, {"disable": True}

    if alias not in _BACKENDS:
        error = _CUSTOM_ERRORS.get(alias, f"wrong backend alias {alias}")
        raise BackendNotAvailableError(error)
    backend_class, pass_uri = _BACKENDS[alias]
    if pass_uri:
        params["address"] = url.split("?")[0]
    return backend_class, params


def serialize_params(params: Dict[str, str]) -> Dict[str, Union[str, int, bool, float]]:
    new_params = {}
    bool_keys = ("safe", "enable", "disable", "client_side")
    true_values = (
        "1",
        "true",
    )
    for key, value in params.items():
        _value: Union[str, int, bool, float]
        if key.lower() in bool_keys:
            _value = value.lower() in true_values
        elif value.isdigit():
            _value = int(value)
        else:
            try:
                _value = float(value)
            except ValueError:
                _value = value
        new_params[key.lower()] = _value
    return new_params



"""
Client side cache is hybrid of mem and redis cache
in theory local cache should be consistence
GET:
-> IN mem cache -> Y -> return
                -> N -> in redis cache -> Y -> store in mem cache -> return
                                       -> N -> compete -> store in mem and in redis -> notify others by channel to invalidate  # noqa: E501

INVALIDATE:

    Redis client side cache cons:
        - if client set he didnt receive message (so we can't update local cache on set without get on redis)
        - message only for first set (2 and after will miss) (solve by request resource after message
        - no control
        - redis >= 6
    Redis client side cache pros:
        + mem cache without ttl
        + no trash

Redis client side caching with non broadcast option weed, with pool of connections it is hard to process connection
lifetime and subscribe for get requests also if we set some value with ttl every client who get value from redis can
also request a ttl and store in local cache with ttl but we steal should know if someone overwrite value or delete it
Broadcasting mode is more useful as we can subscribe for all keys with prefix and invalidate key

https://engineering.redislabs.com/posts/redis-assisted-client-side-caching-in-python/
https://redis.io/topics/client-side-caching
"""

import asyncio
import logging
from typing import Any, AsyncIterator, Mapping, Optional, Tuple

from redis.exceptions import ConnectionError as RedisConnectionError

from .memory import Memory
from .redis import Redis

_REDIS_INVALIDATE_CHAN = "__redis__:invalidate"
_empty = object()
_empty_in_redis = object()  # set when we know that key not in redis
_RECONNECT_WAIT = 10
_DEFAULT_PREFIX = "cashews:"
BCAST_ON = "CLIENT TRACKING on REDIRECT {client_id} BCAST PREFIX {prefix}"
logger = logging.getLogger(__name__)


class BcastClientSide(Redis):
    """
    Cache backend with redis as main storage and client side mem storage that invalidated by
    redis channel for client-side-caching.

    Subscribe with broadcasting by prefix for invalidate by redis>=6
    https://redis.io/topics/client-side-caching
    """

    def __init__(self, *args: Any, local_cache=None, client_side_prefix: str = _DEFAULT_PREFIX, **kwargs: Any) -> None:
        self._local_cache = Memory() if local_cache is None else local_cache
        self._prefix = client_side_prefix
        self._recently_update = Memory(size=500, check_interval=5)
        self._listen_task = None
        self._listen_started = asyncio.Event()
        self.__listen_stop = asyncio.Event()
        super().__init__(*args, **kwargs)

    async def init(self):
        self._listen_started = asyncio.Event()
        self.__listen_stop = asyncio.Event()
        await self._local_cache.init()
        await self._recently_update.init()
        await super().init()
        self.__is_init = False
        self._listen_task = asyncio.create_task(self._listen_invalidate_forever())
        try:
            await asyncio.wait_for(self._listen_started.wait(), timeout=2)
        except TimeoutError:
            if self._listen_task.done():
                raise self._listen_task.exception()
            self._listen_task.cancel()
            raise
        self.__is_init = True

    async def _mark_as_recently_updated(self, key: str):
        await self._recently_update.set(key, True, expire=5)

    def _remove_prefix(self, key: str) -> str:
        return key[len(self._prefix) :]

    def _add_prefix(self, key: str) -> str:
        return self._prefix + key

    async def _listen_invalidate_forever(self):
        while not self.__listen_stop.is_set():
            try:
                await self._listen_invalidate()
            except (RedisConnectionError, ConnectionRefusedError):
                logger.error("broken connection with redis. Clearing client side storage")
                self._listen_started.clear()
                await self._local_cache.clear()
                await asyncio.sleep(_RECONNECT_WAIT)
            except Exception:
                self._listen_started.clear()
                await self._local_cache.clear()
                raise

    async def _get_channel(self):
        pubsub = self._client.pubsub()
        await pubsub.execute_command(b"CLIENT", b"ID")
        client_id = await pubsub.parse_response()
        await pubsub.execute_command(*BCAST_ON.format(client_id=client_id, prefix=self._prefix).encode().split())
        await pubsub.parse_response()
        await pubsub.subscribe(_REDIS_INVALIDATE_CHAN)
        return pubsub

    async def _listen_invalidate(self):
        channel = await self._get_channel()
        self._listen_started.set()
        await self._local_cache.clear()
        while not self.__listen_stop.is_set():
            message = await channel.get_message(ignore_subscribe_messages=True, timeout=0.1)
            if message is None or "data" not in message:
                continue
            if message["data"] is None:  # flushdb
                logger.debug("flush: clear local cache")
                await self._local_cache.clear()
                continue
            for key in message["data"]:
                key = self._remove_prefix(key.decode())
                if not await self._recently_update.get(key):
                    logger.debug("invalidate the key %s", key)
                    await self._local_cache.delete(key)
                else:
                    logger.debug("the key `%s`: recently update", key)
                    await self._recently_update.delete(key)

    async def get(self, key: str, default: Any = None) -> Any:
        if self._listen_started.is_set():
            value = await self._local_cache.get(key, default=_empty)
            if value is _empty_in_redis:
                return default
            if value is not _empty:
                return value
        value = await super().get(self._add_prefix(key), default=_empty)
        if value is not _empty:
            await self._local_cache.set(key, value)
            return value
        await self._local_cache.set(key, _empty_in_redis)
        return default

    async def set(self, key: str, value: Any, *args: Any, **kwargs: Any) -> Any:
        await self._local_cache.set(key, value, *args, **kwargs)
        await self._mark_as_recently_updated(key)
        return await super().set(self._add_prefix(key), value, *args, **kwargs)

    async def set_many(self, pairs: Mapping[str, Any], expire: Optional[float] = None):
        await self._local_cache.set_many(pairs, expire)
        for key in pairs.keys():
            await self._mark_as_recently_updated(key)
        return await super().set_many({self._add_prefix(key): value for key, value in pairs.items()}, expire=expire)

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:  # type: ignore
        async for key in super().scan(self._add_prefix(pattern), batch_size=batch_size):
            yield self._remove_prefix(key)

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Any, ...]:
        missed_keys = {self._add_prefix(key) for key in keys}
        values = {key: default for key in keys}
        if self._listen_started.is_set():
            for i, value in enumerate(await self._local_cache.get_many(*keys, default=_empty)):
                key = keys[i]
                if value is _empty:
                    continue
                if value is _empty_in_redis:
                    value = default
                values[key] = value
                missed_keys.remove(self._add_prefix(key))
        missed_values = await super().get_many(*missed_keys, default=default)
        missed = dict(zip((self._remove_prefix(key) for key in missed_keys), missed_values))
        for key, value in missed.items():
            if value is not default:
                await self._local_cache.set(key, value)
            else:
                await self._local_cache.set(key, _empty_in_redis)
        return tuple(missed.get(key, value) for key, value in values.items())

    async def get_match(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Tuple[str, bytes]]:  # type: ignore
        cursor = 0
        while True:
            cursor, keys = await self._client.scan(cursor, match=self._add_prefix(pattern), count=batch_size)
            if not keys:
                if not cursor:
                    return
                continue
            keys = [self._remove_prefix(key.decode()) for key in keys]
            values = await self.get_many(*keys, default=_empty)
            for key, value in zip(keys, values):
                if value is not _empty:
                    yield key, value
            if not cursor:
                return

    async def incr(self, key: str) -> int:
        value = await super().incr(self._add_prefix(key))
        if value:
            await self._local_cache.set(key, value)
            await self._mark_as_recently_updated(key)
        return value

    async def delete(self, key: str) -> bool:
        await self._local_cache.set(key, _empty_in_redis)
        return await super().delete(self._add_prefix(key))

    async def delete_many(self, *keys: str):
        _keys = []
        for key in keys:
            _keys.append(self._add_prefix(key))
            await self._local_cache.set(key, _empty_in_redis)
        return await super().delete_many(*_keys)

    async def delete_match(self, pattern: str):
        await self._local_cache.delete_match(pattern)
        return await super().delete_match(self._add_prefix(pattern))

    async def expire(self, key: str, timeout: float):
        local_value = await self._local_cache.get(key, default=_empty)
        if local_value not in (_empty, _empty_in_redis):
            await self._local_cache.expire(key, timeout)
            await self._mark_as_recently_updated(key)
        result = await super().expire(self._add_prefix(key), timeout)
        return result

    async def get_expire(self, key: str) -> int:
        if await self._local_cache.get_expire(key) != -1:
            if self._listen_started.is_set():
                return await self._local_cache.get_expire(key)
        expire = await super().get_expire(self._add_prefix(key))
        await self._local_cache.expire(key, expire)
        return expire

    async def exists(self, key) -> bool:
        if self._listen_started.is_set():
            local_value = await self._local_cache.get(key, default=_empty)
            if local_value not in (_empty, _empty_in_redis):
                return True
        return await super().exists(self._add_prefix(key))

    async def set_lock(self, key: str, value, expire):
        await self._mark_as_recently_updated(key)
        await self._local_cache.set_lock(key, value, expire)
        pexpire = None
        if isinstance(expire, float):
            pexpire = int(expire * 1000)
            expire = None
        return bool(await self._client.set(self._add_prefix(key), value, ex=expire, px=pexpire, nx=True))

    async def unlock(self, key, value):
        await self._local_cache.unlock(key, value)
        return await super().unlock(self._add_prefix(key), value)

    async def get_size(self, key: str) -> int:
        return await super().get_size(self._add_prefix(key))

    async def clear(self):
        await self._local_cache.clear()
        return await super().clear()

    async def close(self):
        self.__listen_stop.set()
        if self._listen_task is not None:
            await self._listen_task
            self._listen_task = None
        await self._local_cache.close()
        await self._recently_update.close()
        await super().close()



import time
from contextvars import ContextVar
from functools import wraps
from typing import Any, Callable, Dict, Tuple

from ._typing import AsyncCallable_T, CacheCondition, CallableCacheCondition, Decorator


def _not_none_store_condition(result, args, kwargs, key=None) -> bool:
    return result is not None


def _store_all(result, args, kwargs, key=None) -> bool:
    return True


NOT_NONE = "not_none"
_ALL_CONDITIONS = {Any, "all", any, "any", None}
_NOT_NONE_CONDITIONS = {NOT_NONE, "skip_none"}


def get_cache_condition(condition: CacheCondition) -> Callable[[Any, Tuple, Dict], bool]:
    if condition in _ALL_CONDITIONS:
        return _store_all
    if condition in _NOT_NONE_CONDITIONS:
        return _not_none_store_condition
    return condition


_spent = ContextVar("spent", default=0.0)


def create_time_condition(limit: float) -> Tuple[CallableCacheCondition, Decorator]:
    def decorator(func: AsyncCallable_T) -> AsyncCallable_T:
        @wraps(func)
        async def _wrapper(*args, **kwargs):
            start = time.perf_counter()
            try:
                return await func(*args, **kwargs)
            finally:
                _spent.set(time.perf_counter() - start)

        return _wrapper

    def condition(result: Any, args: Tuple, kwargs: Dict[str, Any], key: str = "") -> bool:
        return _spent.get() > limit

    return condition, decorator



from contextvars import ContextVar
from typing import TYPE_CHECKING, Set

from ._typing import AsyncCallable_T
from .commands import ALL, Command

if TYPE_CHECKING:  # pragma: no cover
    from .backends.interface import Backend


async def _is_disable_middleware(call: AsyncCallable_T, cmd: Command, backend: "Backend", *args, **kwargs):
    if backend.is_disable(cmd):
        if cmd in (Command.GET, Command.GET_MANY):
            return kwargs.get("default", None)
        return None
    return await call(*args, **kwargs)


class ControlMixin:
    def __init__(self) -> None:
        self.__disable: ContextVar[Set[Command]] = ContextVar(str(id(self)), default=set())

    @property
    def _disable(self) -> Set[Command]:
        return self.__disable.get()

    def _set_disable(self, value: Set[Command]) -> None:
        self.__disable.set(value)

    def is_disable(self, *cmds: Command) -> bool:
        _disable = self._disable
        if not cmds and _disable:
            return True
        for cmd in cmds:
            if cmd in _disable:
                return True
        return False

    def is_enable(self, *cmds: Command) -> bool:
        return not self.is_disable(*cmds)

    def disable(self, *cmds: Command) -> None:
        if not cmds:
            _disable = ALL.copy()
        else:
            _disable = self._disable.copy()
            _disable.update(cmds)
        self._set_disable(_disable)

    def enable(self, *cmds: Command) -> None:
        if not cmds:
            _disable = set()
        else:
            _disable = self._disable.copy()
            _disable -= set(cmds)
        self._set_disable(_disable)



from functools import wraps
from typing import Optional, Union

from .._typing import TTL, AsyncCallable_T, Decorator
from ..backends.interface import _BackendInterface
from ..exceptions import LockedError
from ..key import get_cache_key, get_cache_key_template
from ..ttl import ttl_to_seconds

__all__ = ("locked",)


def locked(
    backend: _BackendInterface,
    key: Optional[str] = None,
    ttl: Optional[TTL] = None,
    max_lock_ttl: int = 10,
    step: Union[float, int] = 0.1,
    prefix: str = "lock",
) -> Decorator:
    """
    Decorator that can help you to solve Cache stampede problem (https://en.wikipedia.org/wiki/Cache_stampede),
    Lock following function calls till first one will be finished
    Can guarantee that one function call for given ttl, if ttl is None

    :param backend: cache backend
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param ttl: duration to lock wrapped function call
    :param max_lock_ttl: default ttl if it not set
    :param step: duration between lock check
    :param prefix: custom prefix for key, default 'lock'
    """

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs)
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            try:
                async with backend.lock(_cache_key, _ttl or max_lock_ttl):
                    return await func(*args, **kwargs)
            except LockedError:
                if not await backend.is_locked(_cache_key, wait=_ttl, step=step):
                    return await func(*args, **kwargs)
                raise

        return _wrap

    return _decor



import asyncio
import re
from datetime import datetime
from typing import Any, AsyncIterator, Dict, List, Mapping, Optional, Tuple

from diskcache import Cache, FanoutCache

from cashews.utils import Bitarray

from .interface import Backend


class DiskCache(Backend):
    def __init__(self, *args, directory=None, shards=8, **kwargs: Any) -> None:
        self.__is_init = False
        self._set_locks: Dict[str, asyncio.Lock] = {}
        self._sharded = shards > 1
        if not self._sharded:
            self._cache = Cache(directory=directory, **kwargs)
        else:
            self._cache = FanoutCache(directory=directory, shards=shards, **kwargs)
        super().__init__()

    async def init(self):
        self.__is_init = True

    async def _run_in_executor(self, call, *args):
        return await asyncio.get_running_loop().run_in_executor(None, call, *args)

    @property
    def is_init(self) -> bool:
        return self.__is_init

    async def close(self):
        self._cache.close()
        self.__is_init = False

    async def set(
        self,
        key: str,
        value: Any,
        expire: Optional[float] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        future = self._run_in_executor(self._set, key, value, expire, exist)
        if exist is not None:
            # we should have async lock until value real set
            lock = self._set_locks.setdefault(key, asyncio.Lock())
            async with lock:
                try:
                    return await future
                finally:
                    self._set_locks.pop(key, None)
        return await future

    def _set(self, key, value, expire=None, exist=None):
        if exist is not None:
            if not self._exists(key) is exist:
                return False
        return self._cache.set(key, value, expire)

    async def set_raw(self, key: str, value: Any, **kwargs: Any):
        return self._cache.set(key, value, **kwargs)

    async def get(self, key: str, default: Optional[Any] = None) -> Any:
        return await self._run_in_executor(self._cache.get, key, default)

    async def get_raw(self, key: str) -> Any:
        return self._cache.get(key)

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Any]:
        return await self._run_in_executor(self._get_many, keys, default)

    def _get_many(self, keys: List[str], default: Optional[Any] = None):
        return tuple(self._cache.get(key, default=default) for key in keys)

    async def set_many(self, pairs: Mapping[str, Any], expire: Optional[float] = None):
        return await self._run_in_executor(self._set_many, pairs, expire)

    def _set_many(self, pairs: Mapping[str, Any], expire: Optional[float] = None):
        for key, value in pairs.items():
            self._set(key, value, expire=expire)

    async def exists(self, key: str) -> bool:
        return await self._run_in_executor(self._exists, key)

    def _exists(self, key: str) -> bool:
        return key in self._cache

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:  # type: ignore
        if not self._sharded:
            for key in await self._run_in_executor(self._scan, pattern):
                yield key

    def _scan(self, pattern: str):
        pattern = pattern.replace("*", ".*")
        regexp = re.compile(pattern)
        for key in self._cache.iterkeys():
            if regexp.fullmatch(key):
                yield key

    async def get_bits(self, key: str, *indexes: int, size: int = 1) -> Tuple[int, ...]:
        value = await self.get(key, default=0)
        array = Bitarray(str(value))
        result = []
        for index in indexes:
            result.append(array.get(index, size))
        return tuple(result)

    async def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1) -> Tuple[int, ...]:
        value = await self.get(key, default=0)
        array = Bitarray(str(value))
        result = []
        for index in indexes:
            array.incr(index, size, by)
            result.append(array.get(index, size))
        await self.set(key, array.to_int())
        return tuple(result)

    async def incr(self, key: str) -> int:
        return await self._run_in_executor(self._cache.incr, key)

    async def delete(self, key: str) -> bool:
        return await self._run_in_executor(self._cache.delete, key)

    async def delete_many(self, *keys: str):
        await self._run_in_executor(self._delete_many, keys)

    def _delete_many(self, keys: List[str]):
        for key in keys:
            self._cache.delete(key)

    async def delete_match(self, pattern: str):
        return await self._run_in_executor(self._delete_match, pattern)

    def _delete_match(self, pattern: str):
        for key in self._scan(pattern):
            self._cache.delete(key)

    async def get_match(self, pattern: str, batch_size: int = None) -> AsyncIterator[Tuple[str, Any]]:
        if not self._sharded:
            for key in await self._run_in_executor(self._scan, pattern):
                yield key, await self._run_in_executor(self._cache.get, key)

    async def expire(self, key: str, timeout: float) -> int:
        return await self._run_in_executor(self._cache.touch, key, timeout)

    async def get_expire(self, key: str) -> int:
        return await self._run_in_executor(self._get_expire, key)

    def _get_expire(self, key: str):
        _, expire = self._cache.get(key, expire_time=True)
        if expire is None:
            return -1
        return round((datetime.utcfromtimestamp(expire) - datetime.utcnow()).total_seconds())

    async def get_size(self, key: str) -> int:
        return -1

    async def ping(self, message: Optional[bytes] = None) -> bytes:
        return message or b"PONG"

    async def clear(self):
        await self._run_in_executor(self._cache.clear)

    async def set_lock(self, key: str, value: Any, expire: float) -> bool:
        return await self.set(key, value, expire=expire, exist=False)

    async def is_locked(
        self,
        key: str,
        wait: Optional[float] = None,
        step: float = 0.1,
    ) -> bool:
        if wait is None:
            return await self.exists(key)
        while wait > 0:
            if not await self.exists(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return await self.exists(key)

    async def unlock(self, key: str, value: Any) -> bool:
        return await self.delete(key)



from ._typing import AsyncCallable_T, AsyncCallableResult_T, Middleware
from .backends.interface import Backend
from .commands import PATTERN_CMDS, Command
from .key import get_call_values
from .utils import get_obj_size


def add_prefix(prefix: str) -> Middleware:
    async def _middleware(
        call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs
    ) -> AsyncCallableResult_T:
        if cmd == Command.GET_MANY:
            return await call(*[prefix + key for key in args])
        call_values = get_call_values(call, args, kwargs)
        if cmd == Command.SET_MANY:
            call_values["pairs"] = {prefix + key: value for key, value in call_values["pairs"].items()}
            return await call(**call_values)

        as_key = "pattern" if cmd in PATTERN_CMDS else "key"
        key = call_values.get(as_key)
        if key:
            call_values[as_key] = prefix + key
            return await call(**call_values)
        return await call(*args, **kwargs)

    return _middleware


def all_keys_lower() -> Middleware:
    async def _middleware(
        call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs
    ) -> AsyncCallableResult_T:
        if cmd == Command.GET_MANY:
            return await call(*[key.lower() for key in args])
        call_values = get_call_values(call, args, kwargs)

        if cmd == Command.SET_MANY:
            call_values["pairs"] = {key.lower(): value for key, value in call_values["pairs"].items()}
            return await call(**call_values)

        as_key = "pattern" if cmd in PATTERN_CMDS else "key"

        key = call_values.get(as_key)
        if key:
            call_values[as_key] = key.lower()
            return await call(**call_values)
        return await call(*args, **kwargs)

    return _middleware


def memory_limit(min_bytes=0, max_bytes=None) -> Middleware:
    async def _middleware(
        call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs
    ) -> AsyncCallableResult_T:
        if cmd != Command.SET:
            return await call(*args, **kwargs)
        call_values = get_call_values(call, args, kwargs)
        value_size = get_obj_size(call_values["value"])
        if max_bytes and value_size > max_bytes or value_size < min_bytes:
            return None
        return await call(*args, **kwargs)

    return _middleware



import logging
from datetime import datetime, timedelta
from functools import wraps
from typing import Optional, Tuple, Type, Union

from ..._typing import TTL, AsyncCallable_T, CallableCacheCondition, Decorator
from ...backends.interface import _BackendInterface
from ...formatter import register_template
from ...key import get_cache_key, get_cache_key_template
from ...ttl import ttl_to_seconds
from .defaults import _empty, context_cache_detect

__all__ = ("soft",)


logger = logging.getLogger(__name__)


def soft(
    backend: _BackendInterface,
    ttl: TTL,
    key: Optional[str] = None,
    soft_ttl: Optional[TTL] = None,
    exceptions: Union[Type[Exception], Tuple[Type[Exception], ...]] = Exception,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "soft",
) -> Decorator:
    """
    Cache strategy that allow to use pre-expiration

    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'early'
    """

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs)
            _soft_ttl = ttl_to_seconds(soft_ttl, *args, **kwargs) or _ttl * 0.33
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            cached = await backend.get(_cache_key, default=_empty)
            if cached is not _empty:
                soft_expire_at, result = cached
                if soft_expire_at > datetime.utcnow():
                    context_cache_detect._set(
                        _cache_key,
                        ttl=_ttl,
                        soft_ttl=_soft_ttl,
                        name="soft",
                        template=_key_template,
                    )
                    return result

            try:
                result = await func(*args, **kwargs)
            except exceptions:
                if cached is not _empty:
                    _, result = cached
                    context_cache_detect._set(
                        _cache_key,
                        ttl=_ttl,
                        soft_ttl=_soft_ttl,
                        name="soft",
                        template=_key_template,
                    )
                    return result
                raise
            else:
                if condition(result, args, kwargs, _cache_key):
                    soft_expire_at = datetime.utcnow() + timedelta(seconds=_soft_ttl)
                    await backend.set(_cache_key, [soft_expire_at, result], expire=_ttl)
                return result

        return _wrap

    return _decor



import asyncio
from functools import wraps
from typing import Optional

from ..._typing import TTL, AsyncCallable_T, CallableCacheCondition, Decorator
from ...backends.interface import _BackendInterface
from ...formatter import register_template
from ...key import get_cache_key, get_cache_key_template
from ...ttl import ttl_to_seconds
from .defaults import _empty, context_cache_detect

__all__ = ("hit",)


def hit(
    backend: _BackendInterface,
    ttl: TTL,
    cache_hits: int,
    update_after: Optional[int] = None,
    key: Optional[str] = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "hit",
) -> Decorator:
    """
    Cache call results and drop cache after given numbers of call 'cache_hits'
    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param cache_hits: number of cache hits till cache will dropped
    :param update_after: number of cache hits after cache will update
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'hit'
    """

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs)
            _cache_key = get_cache_key(func, _key_template, args, kwargs)

            result, hits = await asyncio.gather(
                backend.get(_cache_key, default=_empty),
                backend.incr(_cache_key + ":counter"),
            )
            if hits == 1:
                asyncio.create_task(backend.expire(_cache_key + ":counter", _ttl))
            if result is not _empty and hits and hits <= cache_hits:
                context_cache_detect._set(
                    _cache_key,
                    ttl=_ttl,
                    cache_hits=cache_hits,
                    name="hit",
                    template=_key_template,
                )
                if update_after and hits == update_after:
                    asyncio.create_task(_get_and_save(func, args, kwargs, backend, _cache_key, _ttl, condition))
                return result
            return await _get_and_save(func, args, kwargs, backend, _cache_key, _ttl, condition)

        return _wrap

    return _decor


async def _get_and_save(func, args, kwargs, backend, key, ttl, store):
    result = await func(*args, **kwargs)
    if store(result, args, kwargs, key=key):
        await asyncio.gather(
            backend.delete(key + ":counter"),
            backend.set(key, result, expire=ttl),
        )

    return result



from datetime import timedelta
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Tuple, TypeVar, Union

try:
    from typing import Protocol
except ImportError:  # 3.7 python
    from typing_extensions import Protocol

_TTLTypes = Union[int, float, str, timedelta, None]
TTL = Union[_TTLTypes, Callable[[Any], _TTLTypes]]


class CallableCacheCondition(Protocol):
    def __call__(self, result: Any, args: Tuple, kwargs: Dict[str, Any], key: str = "") -> bool:  # pragma: no cover
        ...


CacheCondition = Union[CallableCacheCondition, str, None]

AsyncCallableResult_T = TypeVar("AsyncCallableResult_T")
AsyncCallable_T = Callable[..., Awaitable[AsyncCallableResult_T]]
Decorator = Callable[..., AsyncCallable_T]

if TYPE_CHECKING:  # pragma: no cover
    from . import Command
    from .backends.interface import Backend


class Middleware(Protocol):
    def __call__(
        self,
        call: AsyncCallable_T,
        cmd: "Command",
        backend: "Backend",
        *args,
        **kwargs,
    ) -> Awaitable[AsyncCallableResult_T]:  # pragma: no cover
        ...



import inspect
from functools import lru_cache
from typing import Any, Callable, Container, Dict, Optional, Tuple

from .exceptions import WrongKeyError
from .formatter import _ReplaceFormatter, default_formatter, template_to_pattern

_KWARGS = "__kwargs__"
_ARGS = "__args__"
_ARGS_KWARGS = (_ARGS, _KWARGS)


def get_cache_key(
    func: Callable,
    template: Optional[str] = None,
    args: Tuple[Any, ...] = (),
    kwargs: Optional[Dict] = None,
) -> str:
    """
    Get cache key name for function (:param func) called with args and kwargs
    if func_args is passed key build with parameters are included in func_args dict or tuple otherwise use all of them
    Used function module and name as prefix if key parameter not passed
    :param func: Target function
    :param args: call positional arguments
    :param kwargs: call keyword arguments
    :return: cache key for call
    """
    kwargs = kwargs or {}
    key_values = _get_call_values(func, args, kwargs)
    _key_template = template or get_cache_key_template(func)
    return template_to_pattern(_key_template, _formatter=default_formatter, **key_values)


def get_func_params(func):
    signature = _get_func_signature(func)
    for param_name, param in signature.parameters.items():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            yield _KWARGS
        elif param.kind == inspect.Parameter.VAR_POSITIONAL:
            yield _ARGS
        else:
            yield param_name


def get_cache_key_template(
    func: Callable, key: Optional[str] = None, prefix: str = "", exclude_parameters: Container = ()
) -> str:
    """
    Get cache key name for function (:param func) called with args and kwargs
    Used function module and name as prefix if key parameter not passed
    :param func: Target function
    :param key: template for key, may contain alias to args or kwargs passed to a call
    :param exclude_parameters: array of `args` and `kwargs` names to exclude from a key
        template (if key parameter not passed)
    :return: cache key template
    """

    if key is None:
        key = generate_key_template(func, exclude_parameters)
    else:
        _check_key_params(key, list(get_func_params(func)))
    if prefix:
        key = f"{prefix}:{key}"
    return key


def generate_key_template(func: Callable, exclude_parameters: Container = ()):
    """
    Generate template for function (:param func) called with args and kwargs
    Used function module and name as prefix if key parameter not passed
    :param func: Target function
    :param exclude_parameters: array of `args` and `kwargs` names to exclude
        from a key template (if key parameter not passed)
    :return: cache key template
    """
    func_params = list(get_func_params(func))
    key_template = f"{func.__module__}:{func.__name__}"
    if func_params and func_params[0] == "self":
        key_template = f"{func.__module__}:{func.__qualname__}"
    for param_name in func_params:
        if param_name in exclude_parameters:
            continue
        if param_name in _ARGS_KWARGS:
            key_template += f":{{{param_name}}}"
        else:
            key_template += f":{param_name}:{{{param_name}}}"
    return key_template


class _Star:
    def __getattr__(self, item):
        return _Star()

    def __getitem__(self, item):
        return _Star()


def _check_key_params(key, func_params):
    func_params = {param: _Star() for param in func_params}
    errors = []

    def _default(name):
        errors.append(name)
        return "*"

    check = _ReplaceFormatter(default=_default)
    check.format(key, **func_params)
    if errors:
        raise WrongKeyError(f"Wrong parameter placeholder '{errors}' in the key ")


def get_call_values(func: Callable, args, kwargs) -> Dict:
    """
    Return dict with arguments and their values for function call with given positional and keywords arguments
    :param func: Target function
    :param args: call positional arguments
    :param kwargs: call keyword arguments
    """
    key_values = {}
    for _key, _value in _get_call_values(func, args, kwargs).items():
        if _key not in _ARGS_KWARGS:
            key_values[_key] = _value
    return key_values


@lru_cache(maxsize=100)
def _get_func_signature(func):
    return inspect.signature(func)


def _get_call_values(func, args, kwargs):
    signature = _get_func_signature(func).bind(*args, **kwargs)
    signature.apply_defaults()
    result = {}
    for _name, _value in signature.arguments.items():
        parameter: inspect.Parameter = signature.signature.parameters[_name]
        if parameter.kind == inspect.Parameter.VAR_KEYWORD:
            result[_KWARGS] = _value
            result.update(_value)
        elif parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            result[_ARGS] = _value
        else:
            result[_name] = _value
    return result


def noself(decor_func):
    def _decor(*args, **kwargs):
        def outer(method):
            if "key" not in kwargs:
                kwargs["key"] = get_cache_key_template(method, exclude_parameters={"self"})
            return decor_func(*args, **kwargs)(method)

        return outer

    return _decor



import hashlib
import hmac
from typing import Any, Mapping, Optional, Tuple, Union

from .exceptions import SignIsMissingError, UnSecureDataError
from .picklers import DEFAULT_PICKLE, get_pickler


def get_signer(digestmod):
    def sign(key: bytes, value: bytes):
        return hmac.new(key, value, digestmod).hexdigest().encode()

    return sign


def simple_sign(key: bytes, value: bytes):
    s = sum(key) + sum(value)
    return f"{s:x}".encode()


class PickleSerializerMixin:
    _digestmods = {
        b"sha1": get_signer(hashlib.sha1),
        b"md5": get_signer(hashlib.md5),
        b"sha256": get_signer(hashlib.sha256),
        b"sum": simple_sign,
    }

    def __init__(
        self,
        *args,
        hash_key: Union[str, bytes, None] = None,
        digestmod: Union[str, bytes, None] = b"md5",
        check_repr: bool = True,
        pickle_type: str = DEFAULT_PICKLE,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._hash_key = _to_bytes(hash_key)
        self._digestmod = _to_bytes(digestmod)
        self._check_repr = check_repr
        self._pickler = get_pickler(pickle_type)

    async def get(self, key: str, default: Any = None):
        return await self._serialize_value(await super().get(key), key, default=default)

    async def _serialize_value(self, value, key, default=None) -> Any:
        try:
            return self._process_value(value, key, default=default)
        except (self._pickler.PickleError, AttributeError):
            return default

    def _process_value(self, value: Union[None, int, bytes], key: str, default=None) -> Any:
        if value is None:
            return default
        if isinstance(value, int):
            return value
        if value.isdigit():
            return int(value)

        if not self._hash_key:
            try:
                value = self._pickler.loads(value)
            except self._pickler.PickleError:
                pass
            else:
                if self._check_repr:
                    repr(value)
                return value
        try:
            value = self._get_value_without_signature(value, key)
        except SignIsMissingError:
            return default
        value = self._pickler.loads(value)
        if self._check_repr:
            repr(value)
        return value

    def _get_value_without_signature(self, value: bytes, key: str) -> bytes:
        try:
            sign, value = value.split(b"_", 1)
        except ValueError as exc:
            raise SignIsMissingError(f"key: {key}") from exc
        if not self._hash_key:
            return value
        sign, digestmod = self._get_sign_and_digestmod(sign)
        expected_sign = self._gen_sign(key, value, digestmod)
        if expected_sign != sign:
            raise UnSecureDataError(f"{expected_sign!r} != {sign!r}")
        return value

    def _get_sign_and_digestmod(self, sign: bytes) -> Tuple[bytes, bytes]:
        digestmod = self._digestmod
        if b":" in sign:
            digestmod, sign = sign.split(b":")
        if digestmod not in self._digestmods:
            raise UnSecureDataError()
        return sign, digestmod

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Any:
        values = []
        for key, value in zip(keys, await super().get_many(*keys, default=default) or [None] * len(keys)):
            values.append(await self._serialize_value(value, key))
        return tuple(values)

    async def set(self, key: str, value: Any, *args: Any, **kwargs: Any):
        if isinstance(value, int) and not isinstance(value, bool):
            return await super().set(key, value, *args, **kwargs)
        value = self._pickler.dumps(value)
        return await super().set(key, self._prepend_sign_to_value(key, value), *args, **kwargs)

    async def set_many(self, pairs: Mapping[str, Any], expire: Optional[float] = None):
        transformed_pairs = {}
        for key, value in pairs.items():
            value = self._pickler.dumps(value)
            transformed_pairs[key] = self._prepend_sign_to_value(key, value)
        return await super().set_many(transformed_pairs, expire=expire)

    def _prepend_sign_to_value(self, key: str, value: bytes) -> bytes:
        sign = self._gen_sign(key, value, self._digestmod)
        if not sign:
            return value
        return self._digestmod + b":" + sign + b"_" + value

    def _gen_sign(self, key: str, value: bytes, digestmod: bytes) -> bytes:
        if self._hash_key is None:
            return b""
        value = key.encode() + value
        return self._digestmods[digestmod](self._hash_key, value)

    def set_raw(self, *args: Any, **kwargs: Any):
        return super().set(*args, **kwargs)

    def get_raw(self, *args: Any, **kwargs: Any):
        return super().get(*args, **kwargs)


def _to_bytes(value: Union[str, bytes, None]) -> Optional[bytes]:
    if value is None:
        return None
    if isinstance(value, str):
        value = value.encode()
    return value



import asyncio
from contextlib import contextmanager
from functools import lru_cache, partial, wraps
from typing import Any, AsyncIterator, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Type, Union

from . import decorators, validation
from ._typing import TTL, AsyncCallable_T, AsyncCallableResult_T, CacheCondition, Middleware
from .backend_settings import settings_url_parse
from .backends.interface import Backend, _BackendInterface
from .cache_condition import create_time_condition, get_cache_condition
from .commands import Command
from .disable_control import _is_disable_middleware
from .exceptions import NotConfiguredError
from .ttl import ttl_to_seconds

try:
    from .backends.client_side import BcastClientSide
except ImportError:
    BcastClientSide = None  # type: ignore

try:
    from .backends.diskcache import DiskCache
except ImportError:
    DiskCache = None  # type: ignore


#  pylint: disable=too-many-public-methods


def _create_auto_init() -> Middleware:
    lock = asyncio.Lock()

    async def _auto_init(
        call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs
    ) -> AsyncCallableResult_T:
        if backend.is_init:
            return await call(*args, **kwargs)
        async with lock:
            if not backend.is_init:
                await backend.init()

        return await call(*args, **kwargs)

    return _auto_init


class Cache(_BackendInterface):
    default_prefix = ""

    def __init__(self, name: str = ""):
        self._backends: Dict[str, Tuple[Backend, Tuple[Middleware, ...]]] = {}  # {key: (backend, middleware)}
        self._default_middlewares: Tuple[Middleware, ...] = (
            _is_disable_middleware,
            _create_auto_init(),
            validation._invalidate_middleware,
        )
        self.name = name
        self._default_fail_exceptions: Tuple[Type[Exception], ...] = (Exception,)

    detect = decorators.context_cache_detect

    def set_default_fail_exceptions(self, *exc: Type[Exception]) -> None:
        self._default_fail_exceptions = exc

    def disable(self, *cmds: Command, prefix: str = "") -> None:
        return self._get_backend(prefix).disable(*cmds)

    def enable(self, *cmds: Command, prefix: str = "") -> None:
        return self._get_backend(prefix).enable(*cmds)

    @contextmanager
    def disabling(self, *cmds: Command, prefix: str = ""):
        self.disable(*cmds, prefix=prefix)
        try:
            yield
        finally:
            self.enable(*cmds, prefix=prefix)

    def is_disable(self, *cmds: Command, prefix: str = ""):
        return self._get_backend(prefix).is_disable(*cmds)

    def is_enable(self, *cmds: Command, prefix: str = ""):
        return not self.is_disable(*cmds, prefix=prefix)

    @lru_cache(maxsize=1000)
    def _get_backend_and_config(self, key: str) -> Tuple[Backend, Tuple[Middleware, ...]]:
        for prefix in sorted(self._backends.keys(), reverse=True):
            if key.startswith(prefix):
                return self._backends[prefix]
        if self.default_prefix not in self._backends:
            raise NotConfiguredError("run `cache.setup(...)` before using cache")
        return self._backends[self.default_prefix]

    def _get_backend(self, key: str) -> Backend:
        backend, _ = self._get_backend_and_config(key)
        return backend

    def setup(self, settings_url: str, middlewares: Tuple = (), prefix: str = default_prefix, **kwargs) -> Backend:
        backend_class, params = settings_url_parse(settings_url)
        params.update(kwargs)

        if "disable" in params:
            disable = params.pop("disable")
        else:
            disable = not params.pop("enable", True)

        backend = backend_class(**params)
        if disable:
            backend.disable()
        self._add_backend(backend, middlewares, prefix)
        return backend

    def _add_backend(self, backend: Backend, middlewares=(), prefix: str = default_prefix):
        self._backends[prefix] = (
            backend,
            self._default_middlewares + middlewares,
        )

    async def init(self, *args, **kwargs):
        if args or kwargs:
            self.setup(*args, **kwargs)
        for backend, _ in self._backends.values():
            await backend.init()

    @property
    def is_init(self) -> bool:
        for backend, _ in self._backends.values():
            if not backend.is_init:
                return False
        return True

    def _with_middlewares(self, cmd: Command, key: str):
        backend, middlewares = self._get_backend_and_config(key)
        return self._with_middlewares_for_backend(cmd, backend, middlewares)

    def _with_middlewares_for_backend(self, cmd: Command, backend, middlewares):
        call = getattr(backend, cmd.value)
        for middleware in middlewares:
            call = partial(middleware, call, cmd, backend)
        return call

    async def set(
        self,
        key: str,
        value: Any,
        expire: Union[float, TTL, None] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        return await self._with_middlewares(Command.SET, key)(
            key=key, value=value, expire=ttl_to_seconds(expire), exist=exist
        )

    async def set_raw(self, key: str, value: Any, **kwargs):
        return await self._with_middlewares(Command.SET_RAW, key)(key=key, value=value, **kwargs)

    async def get(self, key: str, default: Optional[Any] = None) -> Any:
        return await self._with_middlewares(Command.GET, key)(key=key, default=default)

    async def get_raw(self, key: str) -> Any:
        return await self._with_middlewares(Command.GET_RAW, key)(key=key)

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:  # type: ignore
        backend, middlewares = self._get_backend_and_config(pattern)

        async def call(pattern, batch_size):
            return backend.scan(pattern, batch_size=batch_size)

        for middleware in middlewares:
            call = partial(middleware, call, Command.SCAN, backend)
        async for key in (await call(pattern=pattern, batch_size=batch_size)):
            yield key

    async def get_match(  # type: ignore
        self,
        pattern: str,
        batch_size: int = 100,
    ) -> AsyncIterator[Tuple[str, Any]]:
        backend, middlewares = self._get_backend_and_config(pattern)

        async def call(pattern, batch_size):
            return backend.get_match(pattern, batch_size=batch_size)

        for middleware in middlewares:
            call = partial(middleware, call, Command.GET_MATCH, backend)
        async for key, value in (await call(pattern=pattern, batch_size=batch_size)):
            yield key, value

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Any, ...]:
        backends: Dict[Backend, List[str]] = {}
        for key in keys:
            backend = self._get_backend(key)
            backends.setdefault(backend, []).append(key)
        result: Dict[str, Any] = {}
        for _keys in backends.values():
            _values = await self._with_middlewares(Command.GET_MANY, _keys[0])(*_keys, default=default)
            result.update(dict(zip(_keys, _values)))
        return tuple(result.get(key) for key in keys)

    async def set_many(self, pairs: Mapping[str, Any], expire: Union[float, TTL, None] = None):
        backends: Dict[Backend, List[str]] = {}
        for key in pairs:
            backend = self._get_backend(key)
            backends.setdefault(backend, []).append(key)
        for backend, keys in backends.items():
            data = {key: pairs[key] for key in keys}
            await self._with_middlewares(Command.SET_MANY, keys[0])(pairs=data, expire=ttl_to_seconds(expire))

    async def get_bits(self, key: str, *indexes: int, size: int = 1) -> Tuple[int, ...]:
        return await self._with_middlewares(Command.GET_BITS, key)(key, *indexes, size=size)

    async def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1):
        return await self._with_middlewares(Command.INCR_BITS, key)(key, *indexes, size=size, by=by)

    async def incr(self, key: str) -> int:
        return await self._with_middlewares(Command.INCR, key)(key=key)

    async def delete(self, key: str):
        return await self._with_middlewares(Command.DELETE, key)(key=key)

    async def delete_many(self, *keys: str) -> None:
        backends: Dict[Backend, List[str]] = {}
        for key in keys:
            backend = self._get_backend(key)
            backends.setdefault(backend, []).append(key)
        for _keys in backends.values():
            await self._with_middlewares(Command.DELETE_MANY, _keys[0])(*_keys)

    async def delete_match(self, pattern: str):
        return await self._with_middlewares(Command.DELETE_MATCH, pattern)(pattern=pattern)

    async def expire(self, key: str, timeout: TTL):
        return await self._with_middlewares(Command.EXPIRE, key)(key=key, timeout=ttl_to_seconds(timeout))

    async def get_expire(self, key: str):
        return await self._with_middlewares(Command.GET_EXPIRE, key)(key=key)

    async def exists(self, key: str):
        return await self._with_middlewares(Command.EXIST, key)(key=key)

    async def set_lock(self, key: str, value: Any, expire: TTL):
        return await self._with_middlewares(Command.SET_LOCK, key)(key=key, value=value, expire=ttl_to_seconds(expire))

    async def unlock(self, key: str, value: str):
        return await self._with_middlewares(Command.UNLOCK, key)(key=key, value=value)

    async def get_size(self, key: str):
        return await self._with_middlewares(Command.GET_SIZE, key)(key=key)

    async def ping(self, message: Optional[bytes] = None) -> bytes:
        message = b"PING" if message is None else message
        return await self._with_middlewares(Command.PING, message.decode())(message=message)

    async def clear(self):
        for backend, _ in self._backends.values():
            await self._with_middlewares_for_backend(Command.CLEAR, backend, self._default_middlewares)()

    async def close(self):
        for backend, _ in self._backends.values():
            await backend.close()

    async def is_locked(
        self,
        key: str,
        wait: Union[float, None, TTL] = None,
        step: Union[int, float] = 0.1,
    ) -> bool:
        return await self._with_middlewares(Command.IS_LOCKED, key)(key=key, wait=ttl_to_seconds(wait), step=step)

    def _wrap_on(self, decorator_fabric, upper, **decor_kwargs):
        if upper:
            return self._wrap_with_condition(decorator_fabric, **decor_kwargs)
        return self._wrap(decorator_fabric, **decor_kwargs)

    def _wrap(self, decorator_fabric, lock=False, time_condition=None, **decor_kwargs):
        def _decorator(func: AsyncCallable_T) -> AsyncCallable_T:
            if time_condition is not None:
                condition, _decor = create_time_condition(time_condition)
                func = _decor(func)
                decor_kwargs["condition"] = condition

            decorator = decorator_fabric(self, **decor_kwargs)(func)

            @wraps(func)
            async def _call(*args, **kwargs):
                if lock:
                    _locked = decorators.locked(self, key=decor_kwargs.get("key"), ttl=decor_kwargs["ttl"])
                    return await _locked(decorator)(*args, **kwargs)
                else:
                    return await decorator(*args, **kwargs)

            return _call

        return _decorator

    def _wrap_with_condition(self, decorator_fabric, condition, lock=False, time_condition=None, **decor_kwargs):
        def _decorator(func: AsyncCallable_T) -> AsyncCallable_T:
            _condition = condition
            if time_condition is not None:
                _condition, _decor = create_time_condition(time_condition)
                func = _decor(func)
            decorator_fabric(self, **decor_kwargs)(func)  # to register cache templates

            @wraps(func)
            async def _call(*args, **kwargs):
                with decorators.context_cache_detect as detect:

                    def new_condition(result, _args, _kwargs, key):
                        if detect.keys:
                            return False
                        return _condition(result, _args, _kwargs, key=key) if _condition else result is not None

                    decorator = decorator_fabric(self, **decor_kwargs, condition=new_condition)
                    if lock:
                        _locked = decorators.locked(self, key=decor_kwargs.get("key"), ttl=decor_kwargs["ttl"])
                        result = await _locked(decorator(func))(*args, **kwargs)
                    else:
                        result = await decorator(func)(*args, **kwargs)

                return result

            return _call

        return _decorator

    # DecoratorS
    def __call__(
        self,
        ttl: TTL,
        key: Optional[str] = None,
        condition: CacheCondition = None,
        time_condition: Optional[TTL] = None,
        prefix: str = "",
        upper: bool = False,
        lock: bool = False,
    ):
        return self._wrap_on(
            decorators.cache,
            upper,
            lock=lock,
            ttl=ttl,
            key=key,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
        )

    cache = __call__

    def failover(
        self,
        ttl: TTL,
        exceptions: Union[Type[Exception], Iterable[Type[Exception]], None] = None,
        key: Optional[str] = None,
        condition: CacheCondition = None,
        time_condition: Optional[TTL] = None,
        prefix: str = "fail",
    ):
        exceptions = exceptions or self._default_fail_exceptions
        return self._wrap_with_condition(
            decorators.failover,
            ttl=ttl,
            exceptions=exceptions,
            key=key,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
        )

    def early(
        self,
        ttl: TTL,
        key: Optional[str] = None,
        early_ttl: Optional[TTL] = None,
        condition: CacheCondition = None,
        time_condition: Optional[TTL] = None,
        prefix: str = "early",
        upper: bool = False,
    ):
        return self._wrap_on(
            decorators.early,
            upper,
            ttl=ttl,
            key=key,
            early_ttl=early_ttl,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
        )

    def soft(
        self,
        ttl: TTL,
        key: Optional[str] = None,
        soft_ttl: Optional[TTL] = None,
        exceptions: Union[Type[Exception], Tuple[Type[Exception], ...]] = Exception,
        condition: CacheCondition = None,
        time_condition: Optional[TTL] = None,
        prefix: str = "soft",
        upper: bool = False,
    ):
        return self._wrap_on(
            decorators.soft,
            upper,
            ttl=ttl,
            key=key,
            soft_ttl=ttl_to_seconds(soft_ttl),
            exceptions=exceptions,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
        )

    def hit(
        self,
        ttl: TTL,
        cache_hits: int,
        update_after: int = 0,
        key: Optional[str] = None,
        condition: CacheCondition = None,
        time_condition: Optional[TTL] = None,
        prefix: str = "hit",
        upper: bool = False,
    ):
        return self._wrap_on(
            decorators.hit,
            upper,
            ttl=ttl,
            cache_hits=cache_hits,
            update_after=update_after,
            key=key,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
        )

    def dynamic(
        self,
        ttl: TTL = 60 * 60 * 24,
        key: Optional[str] = None,
        condition: CacheCondition = None,
        time_condition: Optional[TTL] = None,
        prefix: str = "dynamic",
        upper: bool = False,
    ):
        return self._wrap_on(
            decorators.hit,
            upper,
            ttl=ttl,
            cache_hits=3,
            update_after=1,
            key=key,
            condition=get_cache_condition(condition),
            time_condition=ttl_to_seconds(time_condition),
            prefix=prefix,
        )

    def invalidate(
        self,
        func,
        args_map: Optional[Dict[str, str]] = None,
        defaults: Optional[Dict] = None,
    ):
        return validation.invalidate(
            backend=self,
            target=func,
            args_map=args_map,
            defaults=defaults,
        )

    invalidate_func = validation.invalidate_func

    def circuit_breaker(
        self,
        errors_rate: int,
        period: TTL,
        ttl: TTL,
        half_open_ttl: TTL = None,
        exceptions: Union[Type[Exception], Tuple[Type[Exception], ...], None] = None,
        key: Optional[str] = None,
        prefix: str = "circuit_breaker",
    ):
        _exceptions = exceptions or self._default_fail_exceptions
        return decorators.circuit_breaker(
            backend=self,
            errors_rate=errors_rate,
            period=ttl_to_seconds(period),
            ttl=ttl_to_seconds(ttl),
            half_open_ttl=ttl_to_seconds(half_open_ttl),
            exceptions=_exceptions,
            key=key,
            prefix=prefix,
        )

    def rate_limit(
        self,
        limit: int,
        period: TTL,
        ttl: Optional[TTL] = None,
        action: Optional[Callable] = None,
        prefix="rate_limit",
    ):  # pylint: disable=too-many-arguments
        return decorators.rate_limit(
            backend=self,
            limit=limit,
            period=period,
            ttl=ttl,
            action=action,
            prefix=prefix,
        )

    def locked(
        self,
        ttl: Optional[TTL] = None,
        key: Optional[str] = None,
        step: Union[int, float] = 0.1,
        prefix: str = "locked",
    ):
        return decorators.locked(
            backend=self,
            ttl=ttl,
            key=key,
            step=step,
            prefix=prefix,
        )

    def bloom(
        self,
        *,
        capacity: int,
        name: Optional[str] = None,
        false_positives: Optional[Union[float, int]] = 1,
        check_false_positive: bool = True,
        prefix: str = "bloom",
    ):
        return decorators.bloom(
            backend=self,
            name=name,
            false_positives=false_positives,
            capacity=capacity,
            check_false_positive=check_false_positive,
            prefix=prefix,
        )

    def dual_bloom(
        self,
        *,
        capacity: int,
        name: Optional[str] = None,
        false: Optional[Union[float, int]] = 1,
        no_collisions: bool = False,
        prefix: str = "dual_bloom",
    ):
        return decorators.dual_bloom(
            backend=self,
            name=name,
            false=false,
            no_collisions=no_collisions,
            capacity=capacity,
            prefix=prefix,
        )



class CacheError(Exception):
    pass


class BackendNotAvailableError(CacheError):
    """For wrong or not available cache alias"""


class NotConfiguredError(CacheError):
    """If cache was not configured"""


class UnsupportedPicklerError(CacheError):
    """Unknown or unsupported pickle type."""


class UnSecureDataError(CacheError):
    """Unsecure data in cache storage"""


class SignIsMissingError(CacheError):
    ...


class WrongKeyError(CacheError):
    """Raised If key template have wrong parameter."""


class LockedError(CacheError):
    """Raised if a key already locked"""


class CacheBackendInteractionError(CacheError):
    """Raised if redis not available and safe is set to false"""


class RateLimitError(CacheError):
    """Raised by @rate_limit if rate limit is reached"""



import asyncio
import re
import time
from collections import OrderedDict
from contextlib import suppress
from typing import Any, AsyncIterator, Mapping, Optional, Tuple

from cashews.utils import Bitarray, get_obj_size

from .interface import Backend

__all__ = ["Memory"]

_missed = object()


class Memory(Backend):
    """
    Inmemory backend lru with ttl
    """

    def __init__(self, size: int = 1000, check_interval: float = 1):
        self.store: OrderedDict = OrderedDict()
        self._check_interval = check_interval
        self.size = size
        self.__is_init = False
        self.__remove_expired_stop = asyncio.Event()
        self.__remove_expired_task = None
        super().__init__()

    async def init(self):
        self.__is_init = True
        self.__remove_expired_stop = asyncio.Event()
        self.__remove_expired_task = asyncio.create_task(self._remove_expired())

    @property
    def is_init(self) -> bool:
        return self.__is_init

    async def _remove_expired(self):
        while not self.__remove_expired_stop.is_set():
            for key in dict(self.store):
                await self.get(key)
            with suppress(TimeoutError):
                await asyncio.wait_for(self.__remove_expired_stop.wait(), self._check_interval)

    async def clear(self):
        self.store = OrderedDict()

    async def set(
        self,
        key: str,
        value: Any,
        expire: Optional[float] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        if exist is not None:
            if not (key in self.store) is exist:
                return False
        self._set(key, value, expire)
        return True

    async def set_raw(self, key: str, value: Any, **kwargs: Any):
        self.store[key] = value

    async def get(self, key: str, default: Optional[Any] = None) -> Any:
        return self._get(key, default=default)

    async def get_raw(self, key: str):
        return self.store.get(key)

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Any, ...]:
        return tuple(self._get(key, default=default) for key in keys)

    async def set_many(self, pairs: Mapping[str, Any], expire: Optional[float] = None):
        for key, value in pairs.items():
            self._set(key, value, expire)

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:  # type: ignore
        pattern = pattern.replace("*", ".*")
        regexp = re.compile(pattern)
        for key in dict(self.store):
            if regexp.fullmatch(key):
                yield key

    async def incr(self, key: str):
        value = int(self._get(key, 0)) + 1  # type: ignore
        self._set(key=key, value=value)
        return value

    async def exists(self, key: str):
        return self._key_exist(key)

    async def delete(self, key: str):
        return self._delete(key)

    def _delete(self, key: str) -> bool:
        if key in self.store:
            del self.store[key]
            return True
        return False

    async def delete_many(self, *keys: str):
        for key in keys:
            self._delete(key)

    async def delete_match(self, pattern: str):
        async for key in self.scan(pattern):
            self._delete(key)

    async def get_match(self, pattern: str, batch_size: int = None) -> AsyncIterator[Tuple[str, Any]]:  # type: ignore
        async for key in self.scan(pattern):
            yield key, self._get(key)

    async def expire(self, key: str, timeout: float):
        if not self._key_exist(key):
            return
        value = self._get(key, default=_missed)
        if value is _missed:
            return
        self._set(key, value, timeout)

    async def get_expire(self, key: str) -> int:
        if key not in self.store:
            return -1
        expire_at, _ = self.store[key]
        return round(expire_at - time.time()) if expire_at is not None else -1

    async def ping(self, message: Optional[bytes] = None) -> bytes:
        return b"PONG" if message in (None, b"PING") else message  # type: ignore[return-value]

    async def get_bits(self, key: str, *indexes: int, size: int = 1) -> Tuple[int, ...]:
        array: Bitarray = self._get(key, default=Bitarray("0"))  # type: ignore
        return tuple(array.get(index, size) for index in indexes)

    async def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1) -> Tuple[int, ...]:
        array: Optional[Bitarray] = self._get(key)
        if array is None:
            array = Bitarray("0")
            self._set(key, array)
        result = []
        for index in indexes:
            array.incr(index, size, by)
            result.append(array.get(index, size))
        return tuple(result)

    def _set(self, key: str, value: Any, expire: Optional[float] = None):
        expire = time.time() + expire if expire else None
        if expire is None and key in self.store:
            expire, _ = self.store[key]
        self.store[key] = (expire, value)
        self.store.move_to_end(key)
        if len(self.store) > self.size:
            self.store.popitem(last=False)

    def _get(self, key: str, default: Optional[Any] = None) -> Optional[Any]:
        if key not in self.store:
            return default
        self.store.move_to_end(key)
        expire_at, value = self.store[key]
        if expire_at and expire_at < time.time():
            self._delete(key)
            return default
        return value

    def _key_exist(self, key):
        return self._get(key, default=_missed) is not _missed

    async def set_lock(self, key: str, value: Any, expire: float) -> bool:
        return await self.set(key, value, expire=expire, exist=False)

    async def is_locked(self, key: str, wait: Optional[float] = None, step: float = 0.1) -> bool:
        if wait is None:
            return self._key_exist(key)
        while wait > 0:
            if not self._key_exist(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return self._key_exist(key)

    async def unlock(self, key: str, value: Any) -> bool:
        return self._delete(key)

    async def get_size(self, key: str) -> int:
        if key in self.store:
            return get_obj_size(self.store[key])
        return 0

    async def close(self):
        self.__remove_expired_stop.set()
        if self.__remove_expired_task:
            await self.__remove_expired_task
            self.__remove_expired_task = None
        self.__is_init = False



import asyncio
from typing import Any, AsyncIterator, Dict, Mapping, Optional, Tuple, Type, Union

from redis.asyncio import BlockingConnectionPool

from ..interface import Backend
from .client import Redis, SafeRedis

_UNLOCK = """
if redis.call("get",KEYS[1]) == ARGV[1] then
    return redis.call("del",KEYS[1])
else
    return 0
end
"""
_empty = object()
# pylint: disable=arguments-differ
# pylint: disable=abstract-method


class _Redis(Backend):
    _client: Union[Redis, SafeRedis]
    _client_class: Union[Type[Redis], Type[SafeRedis]]

    def __init__(self, address, safe: bool = True, **kwargs: Any) -> None:
        kwargs.pop("local_cache", None)
        kwargs.pop("prefix", None)
        kwargs.setdefault("client_name", "cashews")
        kwargs.setdefault("health_check_interval", 10)
        kwargs.setdefault("max_connections", 10)
        kwargs.setdefault("socket_keepalive", True)
        kwargs.setdefault("retry_on_timeout", False)
        kwargs.setdefault("socket_timeout", 0.1)
        kwargs["decode_responses"] = False

        self._pool_class = kwargs.pop("connection_pool_class", BlockingConnectionPool)
        if self._pool_class == BlockingConnectionPool:
            kwargs["timeout"] = kwargs.pop("wait_for_connection_timeout", 0.1)
        self._sha: Dict[str, Any] = {}
        if not safe:
            self._client_class = Redis
        else:
            self._client_class = SafeRedis
        self._kwargs = kwargs
        self._address = address
        self.__is_init = False
        super().__init__()

    @property
    def is_init(self) -> bool:
        return self.__is_init

    async def init(self):
        self._client = self._client_class(connection_pool=self._pool_class.from_url(self._address, **self._kwargs))
        await self._client.initialize()
        self.__is_init = True

    async def clear(self):
        return await self._client.flushdb()

    async def set(self, key: str, value: Any, expire: Optional[float] = None, exist=None) -> bool:
        nx = xx = None
        if exist is True:
            xx = True
        elif exist is False:
            nx = True
        pexpire = None
        if isinstance(expire, float):
            pexpire = int(expire * 1000)
            expire = None
        return bool(await self._client.set(key, value, ex=expire, px=pexpire, nx=nx, xx=xx))

    async def set_many(self, pairs: Mapping[str, Any], expire: Optional[float] = None):
        await self._client.mset(pairs)
        if expire is not None:
            async with self._client.pipeline(transaction=True) as pipe:
                for key in pairs.keys():
                    await pipe.pexpire(key, int(expire * 1000))
                await pipe.execute()

    async def get_expire(self, key: str) -> int:
        return await self._client.ttl(key)

    async def expire(self, key: str, timeout: float):
        return await self._client.expire(key, int(timeout))

    async def set_lock(self, key: str, value, expire: float) -> bool:
        pexpire = None
        if isinstance(expire, float):
            pexpire = int(expire * 1000)
            expire = None
        return bool(await self._client.set(key, value, ex=expire, px=pexpire, nx=True))

    async def is_locked(
        self,
        key: str,
        wait: Optional[float] = None,
        step: float = 0.1,
    ) -> bool:
        if wait is None:
            return await self.exists(key)
        async with self._client.client() as conn:
            while wait > 0.0:
                if not await conn.exists(key):
                    return False
                wait -= step
                await asyncio.sleep(step)
        return True

    async def unlock(self, key: str, value: Any) -> bool:
        if "UNLOCK" not in self._sha:
            self._sha["UNLOCK"] = await self._client.script_load(_UNLOCK.replace("\n", " "))
        return await self._client.evalsha(self._sha["UNLOCK"], 1, key, value)

    async def delete(self, key: str) -> bool:
        return bool(await self._client.unlink(key))

    async def exists(self, key: str) -> bool:
        return bool(await self._client.exists(key))

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:  # type: ignore
        cursor = 0
        while True:
            cursor, keys = await self._client.scan(cursor, match=pattern, count=batch_size)
            for key in keys:
                yield key.decode()
            if not cursor:
                return

    async def delete_many(self, *keys: str):
        await self._client.delete(*keys)

    async def delete_match(self, pattern: str):
        if "*" not in pattern:
            await self._client.unlink(pattern)
            return
        cursor = 0
        while True:
            cursor, keys = await self._client.scan(cursor, match=pattern, count=100)
            if not keys:
                if not cursor:
                    return
                continue
            await self._client.unlink(keys[0], *keys[1:])

    async def get_match(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Tuple[str, Any]]:  # type: ignore
        cursor = 0
        while True:
            cursor, keys = await self._client.scan(cursor, match=pattern, count=batch_size)
            if not keys:
                if not cursor:
                    return
                continue
            keys = [key.decode() for key in keys]
            values = await self.get_many(*keys, default=_empty)
            for key, value in zip(keys, values):
                if value is not _empty:  # key can be deleted after scan
                    yield key, value
            if not cursor:
                return

    async def get_size(self, key: str) -> int:
        size = await self._client.memory_usage(key) or 0
        return int(size)

    async def get(self, key: str, default=None) -> Any:
        return await self._client.get(key)

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Optional[Any], ...]:
        if not keys:
            return tuple()
        values = await self._client.mget(keys[0], *keys[1:])
        if not values:
            return tuple()
        return tuple(value if value is not None else default for value in values)

    async def incr(self, key: str) -> int:
        return await self._client.incr(key)

    async def get_bits(self, key: str, *indexes: int, size: int = 1):
        """
        https://redis.io/commands/bitfield
        """
        bitops = self._client.bitfield(key)
        for index in indexes:
            bitops.get(fmt=f"u{size}", offset=f"#{index}")
        return tuple(await bitops.execute() or [])

    async def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1) -> Tuple[int, ...]:
        bitops = self._client.bitfield(key)
        for index in indexes:
            bitops.incrby(fmt=f"u{size}", offset=f"#{index}", increment=by, overflow="SAT")
        return tuple(await bitops.execute())

    async def ping(self, message: Optional[bytes] = None) -> bytes:
        pong = await self._client.ping()
        if pong and message:
            return message
        return b"PONG"

    async def set_raw(self, key: str, value: Any, **kwargs: Any):
        return await self._client.set(key, value, **kwargs)

    async def get_raw(self, key: str) -> Any:
        return await self._client.get(key)

    async def close(self):
        await self._client.close()
        self.__is_init = False



import asyncio
from contextlib import contextmanager
from contextvars import ContextVar
from functools import wraps
from typing import Any, Callable, Dict, Optional, Union

from ._typing import AsyncCallable_T
from .backends.interface import _BackendInterface
from .commands import RETRIEVE_CMDS, Command
from .formatter import get_templates_for_func, template_to_pattern
from .key import get_call_values, get_func_params


async def invalidate_func(backend: _BackendInterface, func, kwargs: Optional[Dict] = None) -> None:
    values = {**{param: "*" for param in get_func_params(func)}, **kwargs}
    for template in get_templates_for_func(func):
        del_template = template_to_pattern(template, **values)
        await backend.delete_match(del_template)


def invalidate(
    backend: _BackendInterface,
    target: Union[str, Callable],
    args_map: Optional[Dict[str, str]] = None,
    defaults: Optional[Dict[str, Any]] = None,
):
    args_map = args_map or {}
    defaults = defaults or {}

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        @wraps(func)
        async def _wrap(*args, **kwargs):
            result = await func(*args, **kwargs)
            _args = get_call_values(func, args, kwargs)
            _args.update(defaults)
            for source, dest in args_map.items():
                if dest in _args:
                    _args[source] = _args.pop(dest)
            if callable(target):
                asyncio.create_task(invalidate_func(backend, target, _args))
            else:
                key = target.format(**{k: str(v) if v is not None else "" for k, v in _args.items()})
                asyncio.create_task(backend.delete_match(key))
            return result

        return _wrap

    return _decor


_INVALIDATE_FURTHER = ContextVar("invalidate", default=False)


@contextmanager
def invalidate_further():
    _INVALIDATE_FURTHER.set(True)
    try:
        yield
    finally:
        _INVALIDATE_FURTHER.set(False)


async def _aiter(num=0):  # pragma: no cover
    """A trick for typing"""
    for i in range(num):
        yield i


async def _invalidate_middleware(call, cmd: Command, backend: _BackendInterface, *args, **kwargs):
    if _INVALIDATE_FURTHER.get() and cmd in RETRIEVE_CMDS:
        if "key" in kwargs:
            asyncio.create_task(backend.delete(kwargs["key"]))
            return kwargs.get("default")
        if cmd == Command.GET_MATCH:
            asyncio.create_task(backend.delete_match(kwargs["pattern"]))
            return _aiter()
        if cmd == Command.GET_MANY:
            asyncio.create_task(backend.delete_many(*args))
            return ()
    return await call(*args, **kwargs)



from datetime import timedelta
from typing import Union

from cashews._typing import TTL


def ttl_to_seconds(ttl: Union[float, None, TTL], *args, **kwargs) -> Union[int, None, float]:
    timeout = ttl(*args, **kwargs) if callable(ttl) else ttl
    if isinstance(timeout, timedelta):
        return timeout.total_seconds()
    if isinstance(timeout, str):
        return _ttl_from_str(timeout)
    return timeout


_STR_TO_DELTA = {
    "h": timedelta(hours=1),
    "m": timedelta(minutes=1),
    "s": timedelta(seconds=1),
    "d": timedelta(days=1),
}


def _ttl_from_str(ttl: str) -> int:
    result = 0
    mul = ""
    for char in ttl.strip().lower():
        if char.isdigit():
            mul += char
        elif char in _STR_TO_DELTA:
            result += int(mul) * int(_STR_TO_DELTA[char].total_seconds())
            mul = ""
        else:
            raise ValueError(f"ttl '{ttl}' has wrong string representation")
    if mul != "" and not result:
        return int(mul)
    return result



import pickle
from typing import Any

from .exceptions import UnsupportedPicklerError

_SQLALC_PICKLE = True
try:
    from sqlalchemy.ext import serializer as sqlalchemy_pickle
except ImportError:
    _SQLALC_PICKLE = False
    sqlalchemy_pickle = pickle

_DILL_PICKLE = True
try:
    import dill
except ImportError:
    _DILL_PICKLE = False
    dill = pickle


class Pickler:
    PickleError = pickle.PickleError

    @staticmethod
    def loads(value: bytes) -> Any:
        return pickle.loads(value, fix_imports=False, encoding="bytes")

    @staticmethod
    def dumps(value: Any) -> bytes:
        return pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL, fix_imports=False)


class SQLAlchemyPickler(Pickler):
    @staticmethod
    def loads(value: bytes) -> Any:
        return sqlalchemy_pickle.loads(
            value,
        )

    @staticmethod
    def dumps(value: Any) -> bytes:
        return sqlalchemy_pickle.dumps(value)


class DillPickler(Pickler):
    @staticmethod
    def loads(value: bytes) -> Any:
        return dill.loads(value)

    @staticmethod
    def dumps(value: Any) -> bytes:
        return dill.dumps(value)


DEFAULT_PICKLE = "default"
_picklers = {
    DEFAULT_PICKLE: Pickler,
    "sqlalchemy": SQLAlchemyPickler,
    "dill": DillPickler,
}


def get_pickler(name: str):
    if name not in _picklers:
        raise UnsupportedPicklerError()

    if name == "sqlalchemy" and not _SQLALC_PICKLE:
        raise UnsupportedPicklerError()

    if name == "dill" and not _DILL_PICKLE:
        raise UnsupportedPicklerError()

    return _picklers[name]
